@@ -1,3 +1,4 @@
+/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
  *
  *  BlueZ - Bluetooth protocol stack for Linux
@@ -32,22 +33,22 @@
 #include <glib/gi18n.h>
 #include <math.h>
 #include <gdk/gdkkeysyms.h>
-#include <nbtk/nbtk-gtk.h>
+#include <mx/mx-gtk.h>
 
 #include "bluetooth-client.h"
 #include "bluetooth-client-private.h"
 #include "bluetooth-chooser.h"
 #include "bluetooth-chooser-private.h"
-#include "bluetooth-killswitch.h"
+#include "bluetooth-powerswitch.h"
 #include "bluetooth-plugin-manager.h"
 #include "bluetooth-filter-widget.h"
 #include "bluetooth-agent.h"
 #include "gnome-bluetooth-enum-types.h"
-#include "bling-spinner.h"
 
 #include "pin.h"
 
 #include "mux-cell-renderer-text.h"
+#include "mux-banner.h"
 #include "koto-cell-renderer-pixbuf.h"
 
 #include "moblin-panel.h"
@@ -60,18 +61,24 @@ G_DEFINE_TYPE (MoblinPanel, moblin_panel, GTK_TYPE_HBOX)
 struct _MoblinPanelPrivate
 {
 	/* Bluetooth objects for interacting with BlueZ */
-	BluetoothKillswitch *killswitch;
+	BluetoothPowerswitch *powerswitch;
 	BluetoothClient *client;
 	BluetoothAgent *agent;
 	GtkTreeModel *chooser_model;
 
 	/* Page widgets that need to be "globally" accessible */
 	GtkWidget *power_switch;
+	guint      visible_countdown; /* seconds */
+	GtkWidget *visible_label;
+	GtkWidget *visible_button;
 	GtkWidget *notebook;
 	GtkWidget *label_pin_help;
 	GtkWidget *label_pin;
 	GtkWidget *label_ssp_pin_help;
 	GtkWidget *label_ssp_pin;
+	GtkWidget *summary_frame;
+	GtkWidget *extra_config_label;
+	GtkWidget *extra_config_vbox;
 	GtkWidget *label_failure;
 	GtkWidget *chooser;
 	GtkWidget *display;
@@ -107,6 +114,7 @@ struct _MoblinPanelPrivate
 };
 
 #define CONNECT_TIMEOUT 3.0
+#define DISCOVER_TIMEOUT (2 * 60)
 #define AGENT_PATH "/org/bluez/agent/moblin"
 
 typedef struct {
@@ -121,11 +129,13 @@ typedef enum {
 	PAGE_SETUP,
 	PAGE_SSP_SETUP,
 	PAGE_CONNECTING,
+	PAGE_SUMMARY,
 	PAGE_FAILURE
 } MoblinPages;
 
 enum {
 	STATUS_CONNECTING,
+        REQUEST_FOCUS,
 	LAST_SIGNAL
 };
 
@@ -138,19 +148,120 @@ static void update_random_pincode (MoblinPanel *self);
 static void create_selected_device (MoblinPanel *self);
 
 static void
-power_switch_toggled_cb (NbtkGtkLightSwitch *light_switch,
+power_switch_toggled_cb (MxGtkLightSwitch *light_switch,
 			 gboolean 	     state,
                          gpointer            user_data)
 {
 	MoblinPanelPrivate *priv = MOBLIN_PANEL_GET_PRIVATE(user_data);
 
 	if (state == TRUE) {
-		bluetooth_killswitch_set_state (priv->killswitch, KILLSWITCH_STATE_UNBLOCKED);
+		bluetooth_powerswitch_set_state (priv->powerswitch, POWERSWITCH_STATE_ON);
 	} else {
-		bluetooth_killswitch_set_state (priv->killswitch, KILLSWITCH_STATE_SOFT_BLOCKED);
+		bluetooth_powerswitch_set_state (priv->powerswitch, POWERSWITCH_STATE_OFF);
 	}
 }
 
+static char *
+totem_time_to_string_text (gint64 msecs)
+{
+	char *secs, *mins, *hours, *string;
+	int sec, min, hour, _time;
+
+	_time = (int) (msecs / 1000);
+	sec = _time % 60;
+	_time = _time - sec;
+	min = (_time % (60*60)) / 60;
+	_time = _time - (min * 60);
+	hour = _time / (60*60);
+
+	hours = g_strdup_printf (ngettext ("%d hour", "%d hours", hour), hour);
+
+	mins = g_strdup_printf (ngettext ("%d minute",
+					  "%d minutes", min), min);
+
+	secs = g_strdup_printf (ngettext ("%d second",
+					  "%d seconds", sec), sec);
+
+	if (hour > 0)
+	{
+		/* hour:minutes:seconds */
+		string = g_strdup_printf (_("%s %s %s"), hours, mins, secs);
+	} else if (min > 0) {
+		if (sec) {
+			/* minutes:seconds */
+			string = g_strdup_printf (_("%s %s"), mins, secs);
+		} else {
+			/* minutes */
+			string = g_strdup_printf (_("%s"), mins);
+		}
+	} else if (sec > 0) {
+		/* seconds */
+		string = g_strdup_printf (_("%s"), secs);
+	} else {
+		/* 0 seconds */
+		string = g_strdup (_("0 seconds"));
+	}
+
+	g_free (hours);
+	g_free (mins);
+	g_free (secs);
+
+	return string;
+}
+
+static gboolean
+update_visible (gpointer data)
+{
+	MoblinPanel *panel = MOBLIN_PANEL (data);
+	MoblinPanelPrivate *priv = MOBLIN_PANEL_GET_PRIVATE (panel);
+	char *label, *time;
+
+	if (priv->visible_countdown > 0) {
+		gtk_widget_hide (priv->visible_button);
+		gtk_widget_show (priv->visible_label);
+
+		time = totem_time_to_string_text (priv->visible_countdown * 1000);
+		label = g_strdup_printf (_("Your computer is visible on\n"
+					   "Bluetooth for %s."), time);
+		gtk_label_set_text (GTK_LABEL (priv->visible_label), label);
+		g_free (label);
+		g_free (time);
+
+		priv->visible_countdown -= 5;
+
+		return TRUE;
+	} else {
+		bluetooth_client_set_discoverable (priv->client, FALSE, 0);
+
+		gtk_widget_show (priv->visible_button);
+		gtk_widget_hide (priv->visible_label);
+
+		return FALSE;
+	}
+}
+
+static void
+make_discoverable (MoblinPanel *panel)
+{
+	MoblinPanelPrivate *priv = MOBLIN_PANEL_GET_PRIVATE (panel);
+
+	if (bluetooth_powerswitch_get_state (priv->powerswitch) != POWERSWITCH_STATE_ON)
+		return;
+
+	bluetooth_client_set_discoverable (priv->client, TRUE, DISCOVER_TIMEOUT);
+
+	priv->visible_countdown = DISCOVER_TIMEOUT;
+
+	update_visible (panel);
+	g_timeout_add_seconds (5, update_visible, panel);
+}
+
+static void
+visible_button_cb (GtkButton *button, gpointer user_data)
+{
+	make_discoverable (MOBLIN_PANEL (user_data));
+}
+
 static void
 enable_send_file (MoblinPanel *self)
 {
@@ -165,6 +276,7 @@ enable_send_file (MoblinPanel *self)
 		guint i;
 		const char **uuids;
 
+		g_free (name);
 		bluetooth_chooser_get_selected_device_info (chooser, "uuids", &value);
 
 		uuids = (const char **) g_value_get_boxed (&value);
@@ -181,8 +293,8 @@ enable_send_file (MoblinPanel *self)
 }
 
 static void
-killswitch_state_changed_cb (BluetoothKillswitch *killswitch,
-                             KillswitchState      state,
+powerswitch_state_changed_cb (BluetoothPowerswitch *powerswitch,
+                             PowerswitchState      state,
                              gpointer             user_data)
 {
 	MoblinPanel *self = MOBLIN_PANEL (user_data);
@@ -190,19 +302,22 @@ killswitch_state_changed_cb (BluetoothKillswitch *killswitch,
 
 	g_signal_handlers_block_by_func (priv->power_switch, power_switch_toggled_cb, user_data);
 
-	if (state == KILLSWITCH_STATE_SOFT_BLOCKED) {
-		nbtk_gtk_light_switch_set_active (NBTK_GTK_LIGHT_SWITCH (priv->power_switch),
+	if (state == POWERSWITCH_STATE_OFF) {
+		mx_gtk_light_switch_set_active (MX_GTK_LIGHT_SWITCH (priv->power_switch),
 	                                      FALSE);
 		gtk_widget_set_sensitive (priv->power_switch, TRUE);
+		gtk_widget_set_sensitive (priv->visible_button, FALSE);
 		gtk_widget_set_sensitive (priv->add_new_button, FALSE);
 		gtk_widget_set_sensitive (priv->send_button, FALSE);
-	} else if (state == KILLSWITCH_STATE_UNBLOCKED) {
-		nbtk_gtk_light_switch_set_active (NBTK_GTK_LIGHT_SWITCH (priv->power_switch), TRUE);
+	} else if (state == POWERSWITCH_STATE_ON) {
+		mx_gtk_light_switch_set_active (MX_GTK_LIGHT_SWITCH (priv->power_switch), TRUE);
 		gtk_widget_set_sensitive (priv->power_switch, TRUE);
+		gtk_widget_set_sensitive (priv->visible_button, TRUE);
 		gtk_widget_set_sensitive (priv->add_new_button, TRUE);
 		enable_send_file (self);
-	} else if (state == KILLSWITCH_STATE_HARD_BLOCKED || state == KILLSWITCH_STATE_NO_ADAPTER) {
+	} else if (state == POWERSWITCH_STATE_NO_ADAPTER) {
 		gtk_widget_set_sensitive (priv->power_switch, FALSE);
+		gtk_widget_set_sensitive (priv->visible_button, FALSE);
 		gtk_widget_set_sensitive (priv->add_new_button, FALSE);
 		gtk_widget_set_sensitive (priv->send_button, FALSE);
 	} else {
@@ -242,7 +357,7 @@ send_file_button_clicked_cb (GtkButton *button,
 	GPtrArray *a;
 	GError *err = NULL;
 	guint i;
-	const char *address, *name;
+	char *address, *name;
 
 	address = bluetooth_chooser_get_selected_device (chooser);
 	name = bluetooth_chooser_get_selected_device_name (chooser);
@@ -263,6 +378,9 @@ send_file_button_clicked_cb (GtkButton *button,
 	}
 	g_ptr_array_add (a, NULL);
 
+	g_free (address);
+	g_free (name);
+
 	if (g_spawn_async(NULL, (char **) a->pdata, NULL,
 			  G_SPAWN_SEARCH_PATH, NULL, NULL, NULL, &err) == FALSE) {
 		g_printerr("Couldn't execute command: %s\n", err->message);
@@ -316,42 +434,27 @@ static void
 set_user_pincode (GtkWidget *button, gpointer user_data)
 {
 	MoblinPanelPrivate *priv = MOBLIN_PANEL_GET_PRIVATE (user_data);
-	GSList *list, *l;
+	GtkEntry *entry;
+	const char *pin;
 
-	list = gtk_radio_button_get_group (GTK_RADIO_BUTTON (button));
-	for (l = list; l ; l = l->next) {
-		GtkEntry *entry;
-		GtkWidget *radio;
-		const char *pin;
+	if (! gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (button)))
+		return;
 
-		if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button)))
-			continue;
+	pin = g_object_get_data (G_OBJECT (button), "pin");
+	entry = g_object_get_data (G_OBJECT (button), "entry");
 
-		/* Is it radio_fixed that changed? */
-		radio = g_object_get_data (G_OBJECT (button), "button");
-		if (radio != NULL) {
-			set_user_pincode (radio, user_data);
-			return;
-		}
-
-		pin = g_object_get_data (G_OBJECT (button), "pin");
-		entry = g_object_get_data (G_OBJECT (button), "entry");
-
-		if (entry != NULL) {
-			g_free (priv->user_pincode);
-			priv->user_pincode = g_strdup (gtk_entry_get_text(entry));
-			gtk_dialog_set_response_sensitive (GTK_DIALOG (priv->pin_dialog),
-							   GTK_RESPONSE_ACCEPT,
-							   gtk_entry_get_text_length (entry) >= 1);
-		} else if (pin != NULL) {
-			g_free (priv->user_pincode);
-			if (*pin == '\0')
-				priv->user_pincode = NULL;
-			else
-				priv->user_pincode = g_strdup (pin);
-		}
-
-		break;
+	if (entry != NULL) {
+		g_free (priv->user_pincode);
+		priv->user_pincode = g_strdup (gtk_entry_get_text(entry));
+		gtk_dialog_set_response_sensitive (GTK_DIALOG (priv->pin_dialog),
+						   GTK_RESPONSE_ACCEPT,
+						   gtk_entry_get_text_length (entry) >= 1);
+	} else if (pin != NULL) {
+		g_free (priv->user_pincode);
+		if (*pin == '\0')
+			priv->user_pincode = NULL;
+		else
+			priv->user_pincode = g_strdup (pin);
 	}
 }
 
@@ -397,31 +500,27 @@ pin_options_button_clicked_cb (GtkButton *button,
 		g_object_set_data (G_OBJECT (priv->radio_1111), "pin", "1111");
 		g_object_set_data (G_OBJECT (priv->radio_1234), "pin", "1234");
 		g_object_set_data (G_OBJECT (priv->radio_custom), "entry", priv->entry_custom);
-
-		gtk_dialog_add_button (GTK_DIALOG (priv->pin_dialog), GTK_STOCK_OK, GTK_RESPONSE_ACCEPT);
 	}
 
-	oldpin = priv->user_pincode;
-	priv->user_pincode = NULL;
+	oldpin = g_strdup (priv->user_pincode);
 
 	gtk_window_present (GTK_WINDOW (priv->pin_dialog));
 
-	if (oldpin == NULL)
+	if (priv->user_pincode == NULL) {
 		radio = priv->radio_auto;
-	else if (g_str_equal (oldpin, "0000"))
+	} else if (g_str_equal (priv->user_pincode, "0000")) {
 		radio = priv->radio_0000;
-	else if (g_str_equal (oldpin, "1111"))
+	} else if (g_str_equal (priv->user_pincode, "1111")) {
 		radio = priv->radio_1111;
-	else if (g_str_equal (oldpin, "1234"))
+	} else if (g_str_equal (priv->user_pincode, "1234")) {
 		radio = priv->radio_1234;
-	else
+	} else {
 		radio = priv->radio_custom;
-
+		gtk_entry_set_text (GTK_ENTRY (priv->entry_custom), priv->user_pincode);
+	}
 	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radio), TRUE);
-	if (radio == priv->radio_custom)
-		gtk_entry_set_text (GTK_ENTRY (priv->entry_custom), oldpin);
 
-	if (gtk_dialog_run (GTK_DIALOG (priv->pin_dialog)) != GTK_RESPONSE_ACCEPT) {
+	if (gtk_dialog_run (GTK_DIALOG (priv->pin_dialog)) != GTK_RESPONSE_OK) {
 		g_free (priv->user_pincode);
 		priv->user_pincode = oldpin;
 	} else {
@@ -429,6 +528,8 @@ pin_options_button_clicked_cb (GtkButton *button,
 	}
 
 	gtk_widget_hide (priv->pin_dialog);
+
+        g_signal_emit (self, _signals[REQUEST_FOCUS], 0);
 }
 
 /*
@@ -459,14 +560,14 @@ remove_clicked_cb (GtkCellRenderer *cell, const gchar *path, gpointer user_data)
 	ensure_selection (chooser, path);
 
 	/* Get address */
-	if (bluetooth_chooser_get_selected_device_info (chooser, "address", &value)) {
-		address = g_value_get_string (&value);
-		g_value_unset (&value);
-	}
+	g_assert (bluetooth_chooser_get_selected_device_info (chooser, "address", &value));
+	address = g_value_get_string (&value);
 
 	if (bluetooth_chooser_remove_selected_device (chooser) != FALSE && address) {
 		bluetooth_plugin_manager_device_deleted (address);
 	}
+
+	g_value_unset (&value);
 }
 
 static void
@@ -483,7 +584,6 @@ browse_clicked (GtkCellRenderer *renderer, const gchar *path, gpointer user_data
 	/* Get address */
 	if (bluetooth_chooser_get_selected_device_info (chooser, "address", &value)) {
 		address = g_value_get_string (&value);
-		g_value_unset (&value);
 	}
 
 	if (address == NULL) {
@@ -493,6 +593,8 @@ browse_clicked (GtkCellRenderer *renderer, const gchar *path, gpointer user_data
 			g_printerr("Couldn't execute command: %s\n", cmd);
 		g_free (cmd);
 	}
+
+	g_value_unset (&value);
 }
 
 static void
@@ -601,21 +703,38 @@ connect_device (const gchar *device_path, MoblinPanel *self)
 	}
 }
 
+static void
+container_remove_all (GtkContainer *container)
+{
+	GList *list;
+	GList *iter;
+
+	list = gtk_container_get_children (container);
+	for (iter = list; iter; iter = iter->next)
+		gtk_container_remove (container, GTK_WIDGET (iter->data));
+
+	g_list_free (list);
+}
+
 static void
 set_current_page (MoblinPanel *self, MoblinPages page)
 {
 	MoblinPanelPrivate *priv = MOBLIN_PANEL_GET_PRIVATE (self);
 
 	if (page == PAGE_ADD) {
+		g_object_set (BLUETOOTH_CHOOSER (priv->chooser),
+			      "device-category-filter", BLUETOOTH_CATEGORY_NOT_PAIRED_OR_TRUSTED,
+			      NULL);
+
 		bluetooth_chooser_start_discovery (BLUETOOTH_CHOOSER (priv->chooser));
 	} else {
 		bluetooth_chooser_stop_discovery (BLUETOOTH_CHOOSER (priv->chooser));
 	}
 
 	if (page == PAGE_CONNECTING)
-		bling_spinner_start (BLING_SPINNER (priv->spinner));
+		gtk_spinner_start (GTK_SPINNER (priv->spinner));
 	else
-		bling_spinner_stop (BLING_SPINNER (priv->spinner));
+		gtk_spinner_stop (GTK_SPINNER (priv->spinner));
 
 	if ((page == PAGE_SETUP || page == PAGE_SSP_SETUP || page == PAGE_CONNECTING) && (priv->create_started == FALSE)) {
 		create_selected_device (self);
@@ -649,6 +768,43 @@ set_current_page (MoblinPanel *self, MoblinPages page)
 		gtk_widget_hide (priv->does_not_match_button);
 	}
 
+	if (page == PAGE_SUMMARY) {
+		GValue value = { 0, };
+		const char **uuids;
+		GList *widgets;
+		char *title;
+
+		title = g_strdup_printf (_("Successfully set up new device '%s'"), priv->target_name);
+		set_frame_title (GTK_FRAME (priv->summary_frame), title);
+		g_free (title);
+
+		if (bluetooth_chooser_get_selected_device_info (BLUETOOTH_CHOOSER (priv->chooser), "uuids", &value))
+			uuids = g_value_get_boxed (&value);
+		else
+			uuids = NULL;
+
+		widgets = bluetooth_plugin_manager_get_widgets (priv->target_address, uuids);
+		if (widgets) {
+			GList *iter;
+
+			for (iter = widgets; iter; iter = iter->next)
+				gtk_box_pack_start (GTK_BOX (priv->extra_config_vbox),
+						    GTK_WIDGET (iter->data),
+						    FALSE,
+						    TRUE,
+						    0);
+
+			g_list_free (widgets);
+			gtk_widget_show (priv->extra_config_label);
+			gtk_widget_show_all (priv->extra_config_vbox);
+		} else {
+			gtk_widget_hide (priv->extra_config_label);
+			gtk_widget_hide (priv->extra_config_vbox);
+		}
+
+		g_value_unset (&value);
+	}
+
 	if (page == PAGE_DEVICES) {
 		/* Clean up old state */
 		update_random_pincode (self);
@@ -659,6 +815,7 @@ set_current_page (MoblinPanel *self, MoblinPages page)
 		priv->target_address = NULL;
 		g_free (priv->target_name);
 		priv->target_name = NULL;
+		container_remove_all (GTK_CONTAINER (priv->extra_config_vbox));
 		g_object_set (priv->chooser, "device-type-filter", BLUETOOTH_TYPE_ANY, NULL);
 	}
 
@@ -670,12 +827,14 @@ create_callback (BluetoothClient *client, const gchar *path, const GError *error
 {
 	MoblinPanel *self = MOBLIN_PANEL (user_data);
 	MoblinPanelPrivate *priv = MOBLIN_PANEL_GET_PRIVATE (self);
-	gchar *device_name = bluetooth_chooser_get_selected_device_name (BLUETOOTH_CHOOSER (priv->chooser));
+	gchar *device_name;
 
 	priv->create_started = FALSE;
 
 	if (path == NULL) {
+		device_name = bluetooth_chooser_get_selected_device_name (BLUETOOTH_CHOOSER (priv->chooser));
 		set_failure_message (self, device_name);
+		g_free (device_name);
 		set_current_page (self, PAGE_FAILURE);
 		return;
 	}
@@ -684,7 +843,7 @@ create_callback (BluetoothClient *client, const gchar *path, const GError *error
 
 	connect_device (path, self);
 
-	set_current_page (self, PAGE_DEVICES);
+	set_current_page (self, PAGE_SUMMARY);
 }
 
 static void
@@ -694,6 +853,12 @@ create_selected_device (MoblinPanel *self)
 	const gchar *path = AGENT_PATH;
 	gchar *pin_ret;
 
+	/* Set the filter on the selector, so we can use it to get more
+	 * info later, in page_summary */
+	g_object_set (priv->chooser,
+		      "device-category-filter", BLUETOOTH_CATEGORY_ALL,
+		      NULL);
+
 	pin_ret = get_pincode_for_device (priv->target_type, priv->target_address,
 					priv->target_name, NULL);
 	if (pin_ret != NULL && g_str_equal (pin_ret, "NULL"))
@@ -728,6 +893,7 @@ pair_clicked (GtkCellRenderer *renderer, const gchar *path, gpointer user_data)
 	}
 	if (bluetooth_chooser_get_selected_device_info (chooser, "legacypairing", &value) != FALSE) {
 		legacy_pairing = g_value_get_int (&value);
+		g_value_unset (&value);
 			if (legacy_pairing == -1)
 				legacy_pairing = TRUE;
 	} else {
@@ -735,10 +901,10 @@ pair_clicked (GtkCellRenderer *renderer, const gchar *path, gpointer user_data)
 	}
 
 	g_free (priv->target_address);
-	priv->target_address = g_strdup (address);
+	priv->target_address = address;
 
 	g_free (priv->target_name);
-	priv->target_name = g_strdup (name);
+	priv->target_name = name;
 
 	priv->target_type = type;
 	priv->target_ssp = !legacy_pairing;
@@ -1040,7 +1206,7 @@ create_failure_page (MoblinPanel *self)
 
 	priv = MOBLIN_PANEL_GET_PRIVATE (self);
 
-	page = nbtk_gtk_frame_new ();
+	page = mx_gtk_frame_new ();
 	page_title = gtk_label_new ("");
 	gtk_frame_set_label_widget (GTK_FRAME (page), page_title);
 	set_frame_title (GTK_FRAME (page), _("Device setup failed"));
@@ -1050,7 +1216,9 @@ create_failure_page (MoblinPanel *self)
 	gtk_widget_show (vbox);
 	gtk_container_add (GTK_CONTAINER (page), vbox);
 	priv->label_failure = gtk_label_new ("");
+	gtk_misc_set_alignment (GTK_MISC (priv->label_failure), 0.0, 0.5);
 	gtk_widget_show (priv->label_failure);
+	gtk_box_pack_start (GTK_BOX (vbox), priv->label_failure, FALSE, FALSE, 6);
 
 	hbox = gtk_hbox_new (FALSE, 6);
 	gtk_widget_show (hbox);
@@ -1063,6 +1231,45 @@ create_failure_page (MoblinPanel *self)
 	return page;
 }
 
+static GtkWidget *
+create_summary_page (MoblinPanel *self)
+{
+	MoblinPanelPrivate *priv;
+	GtkWidget *page;
+	GtkWidget *vbox, *hbox;
+	GtkWidget *w;
+
+	priv = MOBLIN_PANEL_GET_PRIVATE (self);
+
+	page = mx_gtk_frame_new ();
+	w = gtk_label_new (NULL);
+	gtk_frame_set_label_widget (GTK_FRAME (page), w);
+	gtk_widget_show (page);
+	priv->summary_frame = page;
+
+	vbox = gtk_vbox_new (FALSE, 12);
+	gtk_widget_show (vbox);
+	gtk_container_add (GTK_CONTAINER (page), vbox);
+	priv->extra_config_label = gtk_label_new (_("Select the additional services you want to use with your device:"));
+	gtk_misc_set_alignment (GTK_MISC (priv->extra_config_label), 0.0, 0.5);
+	gtk_widget_show (priv->extra_config_label);
+	gtk_box_pack_start (GTK_BOX (vbox), priv->extra_config_label, FALSE, FALSE, 6);
+
+	priv->extra_config_vbox = gtk_vbox_new (FALSE, 6);
+	gtk_widget_show (priv->extra_config_vbox);
+	gtk_box_pack_start (GTK_BOX (vbox), priv->extra_config_vbox, TRUE, TRUE, 6);
+
+	hbox = gtk_hbox_new (FALSE, 6);
+	gtk_widget_show (hbox);
+	gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, FALSE, 6);
+	w = gtk_button_new_with_label (_("Done"));
+	gtk_widget_show (w);
+	gtk_box_pack_start (GTK_BOX (hbox), w, FALSE, FALSE, 6);
+	g_signal_connect (w, "clicked", G_CALLBACK (set_device_view), self);
+
+	return page;
+}
+
 static GtkWidget *
 create_setup_page (MoblinPanel *self)
 {
@@ -1073,7 +1280,7 @@ create_setup_page (MoblinPanel *self)
 
 	priv = MOBLIN_PANEL_GET_PRIVATE (self);
 
-	page = nbtk_gtk_frame_new ();
+	page = mx_gtk_frame_new ();
 	page_title = gtk_label_new ("");
 	gtk_frame_set_label_widget (GTK_FRAME (page), page_title);
 	set_frame_title (GTK_FRAME (page), _("Device setup"));
@@ -1110,7 +1317,7 @@ create_ssp_setup_page (MoblinPanel *self)
 
 	priv = MOBLIN_PANEL_GET_PRIVATE (self);
 
-	page = nbtk_gtk_frame_new ();
+	page = mx_gtk_frame_new ();
 	page_title = gtk_label_new ("");
 	gtk_frame_set_label_widget (GTK_FRAME (page), page_title);
 	set_frame_title (GTK_FRAME (page), _("Device setup"));
@@ -1156,8 +1363,10 @@ create_add_page (MoblinPanel *self)
 	GtkWidget *page;
 	GtkWidget *vbox, *hbox;
 	GtkWidget *filter;
-	GtkWidget *frame_title;
+	GtkWidget *frame_vbox;
+	GtkWidget *banner;
 	GtkWidget *frame;
+	GtkWidget *alignment;
 	GtkWidget *back_button;
 	GtkWidget *pin_button;
 	GtkWidget *tree_view;
@@ -1166,18 +1375,34 @@ create_add_page (MoblinPanel *self)
 
 	priv = MOBLIN_PANEL_GET_PRIVATE (self);
 
-	page = gtk_hbox_new (FALSE, 0);
+	page = gtk_hbox_new (FALSE, 8);
+	gtk_container_set_border_width (GTK_CONTAINER (page), 8);
 	gtk_widget_show (page);
 
 	/* Add child widgetry */
 	vbox = gtk_vbox_new (FALSE, 4);
 	gtk_widget_show (vbox);
-	gtk_box_pack_start (GTK_BOX (page), vbox, TRUE, TRUE, 4);
+	gtk_box_pack_start (GTK_BOX (page), vbox, TRUE, TRUE, 0);
 
-	frame = nbtk_gtk_frame_new ();
-	frame_title = gtk_label_new ("");
-	gtk_frame_set_label_widget (GTK_FRAME (frame), frame_title);
-	set_frame_title (GTK_FRAME (frame), _("Devices"));
+	frame = gtk_frame_new (NULL);
+	gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_OUT);
+	gtk_widget_show (frame);
+	gtk_box_pack_start (GTK_BOX (vbox), frame, TRUE, TRUE, 0);
+
+	frame_vbox = gtk_vbox_new (FALSE, 0);
+	gtk_widget_show (frame_vbox);
+	gtk_container_add (GTK_CONTAINER (frame), frame_vbox);
+
+	banner = mux_banner_new (_("Devices"));
+	gtk_widget_show (banner);
+	gtk_box_pack_start (GTK_BOX (frame_vbox), banner, FALSE, FALSE, 0);
+
+	/* BluetoothChooser adds its own alignment on the left, so balance it
+	   out... */
+	alignment = gtk_alignment_new (0.0, 0.0, 1.0, 1.0);
+	gtk_alignment_set_padding (GTK_ALIGNMENT (alignment), 12, 12, 0, 12);
+	gtk_widget_show (alignment);
+	gtk_box_pack_start (GTK_BOX (frame_vbox), alignment, TRUE, TRUE, 0);
 
 	/* Device list */
 	priv->chooser = g_object_new (BLUETOOTH_TYPE_CHOOSER,
@@ -1198,28 +1423,38 @@ create_add_page (MoblinPanel *self)
 	g_signal_connect (cell, "activated", G_CALLBACK (pair_clicked), self);
 
 	gtk_widget_show (priv->chooser);
-	gtk_container_add (GTK_CONTAINER (frame), priv->chooser);
-	gtk_widget_show (frame);
-	gtk_box_pack_start (GTK_BOX (vbox), frame, FALSE, FALSE, 4);
+	gtk_container_add (GTK_CONTAINER (alignment), priv->chooser);
 
 	/* Back button */
 	back_button = gtk_button_new_with_label (_("Back to devices"));
 	gtk_widget_show (back_button);
 	g_signal_connect (back_button, "clicked",
 			G_CALLBACK (set_device_view), self);
-	gtk_box_pack_start (GTK_BOX (vbox), back_button, FALSE, FALSE, 4);
+	gtk_box_pack_start (GTK_BOX (vbox), back_button, FALSE, FALSE, 0);
 
 	/* Right column */
-	frame = nbtk_gtk_frame_new ();
+	frame = gtk_frame_new (NULL);
+	gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_OUT);
 	gtk_widget_show (frame);
-	vbox = gtk_vbox_new (FALSE, 4);
+	gtk_box_pack_start (GTK_BOX (page), frame, FALSE, FALSE, 0);
+
+	frame_vbox = gtk_vbox_new (FALSE, 0);
+	gtk_container_set_border_width (GTK_CONTAINER (frame_vbox), 0);
+	gtk_widget_show (frame_vbox);
+	gtk_container_add (GTK_CONTAINER (frame), frame_vbox);
+
+	banner = mux_banner_new (_("Settings"));
+	gtk_widget_show (banner);
+	gtk_box_pack_start (GTK_BOX (frame_vbox), banner, FALSE, FALSE, 0);
+
+	vbox = gtk_vbox_new (FALSE, 0);
+	gtk_container_set_border_width (GTK_CONTAINER (vbox), 12);
 	gtk_widget_show (vbox);
-	gtk_container_add (GTK_CONTAINER (frame), vbox);
-	gtk_box_pack_start (GTK_BOX (page), frame, FALSE, FALSE, 4);
+	gtk_container_add (GTK_CONTAINER (frame_vbox), vbox);
 
 	hbox = gtk_hbox_new (FALSE, 4);
 	gtk_widget_show (hbox);
-	gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 4);
+	gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
 
 	/* Filter combo */
 	filter = bluetooth_filter_widget_new ();
@@ -1227,14 +1462,13 @@ create_add_page (MoblinPanel *self)
 	bluetooth_filter_widget_set_title (BLUETOOTH_FILTER_WIDGET (filter), _("Only show:"));
 	bluetooth_filter_widget_bind_filter (BLUETOOTH_FILTER_WIDGET (filter),
 					     BLUETOOTH_CHOOSER (priv->chooser));
-	gtk_box_pack_start (GTK_BOX (vbox), filter, FALSE, FALSE, 4);
+	gtk_box_pack_start (GTK_BOX (vbox), filter, FALSE, FALSE, 0);
 
 	/* Button for PIN options file */
 	pin_button = gtk_button_new_with_label (_("PIN options"));
 	gtk_widget_show (pin_button);
-	g_signal_connect (pin_button, "clicked",
-                    G_CALLBACK (pin_options_button_clicked_cb), self);
-	gtk_box_pack_start (GTK_BOX (vbox), pin_button, FALSE, FALSE, 4);
+	g_signal_connect (pin_button, "clicked", G_CALLBACK (pin_options_button_clicked_cb), self);
+	gtk_box_pack_start (GTK_BOX (vbox), pin_button, FALSE, FALSE, 0);
 
 	return page;
 }
@@ -1244,28 +1478,49 @@ create_devices_page (MoblinPanel *self)
 {
 	MoblinPanelPrivate *priv;
 	GtkWidget *page;
-	GtkWidget *frame_title;
-	GtkWidget *vbox, *hbox;
+	GtkWidget *banner;
+	GtkWidget *alignment, *vbox, *vbox2, *hbox;
 	GtkWidget *frame;
 	GtkWidget *power_label;
 	GtkTreeViewColumn *type_column;
 	GtkCellRenderer *cell;
-	KillswitchState switch_state;
 	GtkWidget *tree_view;
+	GtkSizeGroup *size;
 
 	priv = MOBLIN_PANEL_GET_PRIVATE (self);
 
-	page = gtk_hbox_new (FALSE, 0);
+	page = gtk_hbox_new (FALSE, 8);
+	gtk_container_set_border_width (GTK_CONTAINER (page), 8);
 	gtk_widget_show (page);
-	/* Add child widgetry */
-	vbox = gtk_vbox_new (FALSE, 4);
-	gtk_widget_show (vbox);
-	gtk_box_pack_start (GTK_BOX (page), vbox, TRUE, TRUE, 4);
 
-	frame = nbtk_gtk_frame_new ();
-	frame_title = gtk_label_new ("");
-	gtk_frame_set_label_widget (GTK_FRAME (frame), frame_title);
-	set_frame_title (GTK_FRAME (frame), _("Devices"));
+	size = gtk_size_group_new (GTK_SIZE_GROUP_VERTICAL);
+
+	frame = gtk_frame_new (NULL);
+	gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_OUT);
+	gtk_widget_show (frame);
+	vbox = gtk_vbox_new (FALSE, 0);
+	gtk_widget_show (vbox);
+	gtk_container_add (GTK_CONTAINER (frame), vbox);
+	gtk_box_pack_start (GTK_BOX (page), frame, TRUE, TRUE, 0);
+
+	banner = mux_banner_new (_("Devices"));
+	gtk_widget_show (banner);
+	gtk_size_group_add_widget (size, banner);
+	gtk_box_pack_start (GTK_BOX (vbox), banner, FALSE, FALSE, 0);
+
+	/* Add new button */
+	priv->add_new_button = gtk_button_new_with_label (_("Add a new device"));
+	gtk_widget_show (priv->add_new_button);
+	g_signal_connect (priv->add_new_button, "clicked",
+			G_CALLBACK (set_scanning_view), self);
+	gtk_box_pack_end (GTK_BOX (banner), priv->add_new_button, FALSE, FALSE, 0);
+
+	/* BluetoothChooser adds its own alignment on the left, so balance it
+	   out... */
+	alignment = gtk_alignment_new (0.0, 0.0, 1.0, 1.0);
+	gtk_alignment_set_padding (GTK_ALIGNMENT (alignment), 12, 12, 0, 12);
+	gtk_widget_show (alignment);
+	gtk_box_pack_start (GTK_BOX (vbox), alignment, TRUE, TRUE, 0);
 
 	/* Device list */
 	priv->display = g_object_new (BLUETOOTH_TYPE_CHOOSER,
@@ -1308,28 +1563,30 @@ create_devices_page (MoblinPanel *self)
 	g_signal_connect (cell, "activated", G_CALLBACK (remove_clicked_cb), self);
 
 	gtk_widget_show (priv->display);
-	gtk_container_add (GTK_CONTAINER (frame), priv->display);
-	gtk_widget_show (frame);
-	gtk_box_pack_start (GTK_BOX (vbox), frame, FALSE, FALSE, 4);
-
-	/* Add new button */
-	priv->add_new_button = gtk_button_new_with_label (_("Add a new device"));
-	gtk_widget_show (priv->add_new_button);
-	g_signal_connect (priv->add_new_button, "clicked",
-			G_CALLBACK (set_scanning_view), self);
-	gtk_box_pack_start (GTK_BOX (vbox), priv->add_new_button, FALSE, FALSE, 4);
+	gtk_container_add (GTK_CONTAINER (alignment), priv->display);
 
 	/* Right column */
-	frame = nbtk_gtk_frame_new ();
+	frame = gtk_frame_new (NULL);
+	gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_OUT);
 	gtk_widget_show (frame);
-	vbox = gtk_vbox_new (FALSE, 4);
+	vbox = gtk_vbox_new (FALSE, 0);
 	gtk_widget_show (vbox);
 	gtk_container_add (GTK_CONTAINER (frame), vbox);
-	gtk_box_pack_start (GTK_BOX (page), frame, FALSE, FALSE, 4);
+	gtk_box_pack_start (GTK_BOX (page), frame, FALSE, FALSE, 0);
+
+	banner = mux_banner_new (_("Settings"));
+	gtk_size_group_add_widget (size, banner);
+	gtk_widget_show (banner);
+	gtk_box_pack_start (GTK_BOX (vbox), banner, FALSE, FALSE, 0);
+
+	vbox2 = gtk_vbox_new (FALSE, 4);
+	gtk_container_set_border_width (GTK_CONTAINER (vbox2), 12);
+	gtk_widget_show (vbox2);
+	gtk_box_pack_start (GTK_BOX (vbox), vbox2, TRUE, TRUE, 0);
 
 	hbox = gtk_hbox_new (FALSE, 4);
 	gtk_widget_show (hbox);
-	gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 4);
+	gtk_box_pack_start (GTK_BOX (vbox2), hbox, FALSE, FALSE, 4);
 
 	/* Power switch */
 	/* Translators: This string appears next to a toggle switch which controls
@@ -1339,46 +1596,25 @@ create_devices_page (MoblinPanel *self)
 	gtk_widget_show (power_label);
 	gtk_box_pack_start (GTK_BOX (hbox), power_label, FALSE, FALSE, 4);
 
-	priv->power_switch = nbtk_gtk_light_switch_new ();
-	if (priv->killswitch != NULL) {
-		if (bluetooth_killswitch_has_killswitches (priv->killswitch) == FALSE) {
-			gtk_widget_set_sensitive (priv->power_switch, FALSE);
-			gtk_widget_set_sensitive (priv->add_new_button, FALSE);
-		} else {
-			switch_state = bluetooth_killswitch_get_state (priv->killswitch);
-
-			switch (switch_state) {
-				case KILLSWITCH_STATE_UNBLOCKED:
-					nbtk_gtk_light_switch_set_active
-						(NBTK_GTK_LIGHT_SWITCH (priv->power_switch),
-						 TRUE);
-					gtk_widget_set_sensitive (priv->add_new_button, TRUE);
-					gtk_widget_set_sensitive (priv->power_switch, TRUE);
-				break;
-				case KILLSWITCH_STATE_SOFT_BLOCKED:
-					nbtk_gtk_light_switch_set_active
-						(NBTK_GTK_LIGHT_SWITCH (priv->power_switch),
-							FALSE);
-				break;
-				case KILLSWITCH_STATE_HARD_BLOCKED:
-				case KILLSWITCH_STATE_NO_ADAPTER:
-				default:
-					gtk_widget_set_sensitive (priv->power_switch, FALSE);
-					gtk_widget_set_sensitive (priv->add_new_button, FALSE);
-				break;
-			}
-		}
-	} else {
-		gtk_widget_set_sensitive (priv->power_switch, FALSE);
-		gtk_widget_set_sensitive (priv->add_new_button, FALSE);
-	}
-	g_signal_connect  (priv->killswitch, "state-changed",
-			G_CALLBACK (killswitch_state_changed_cb), self);
+	priv->power_switch = mx_gtk_light_switch_new ();
+	g_signal_connect  (priv->powerswitch, "state-changed",
+			G_CALLBACK (powerswitch_state_changed_cb), self);
 	g_signal_connect (priv->power_switch, "switch-flipped",
 			G_CALLBACK (power_switch_toggled_cb), self);
 	gtk_widget_show (priv->power_switch);
 	gtk_box_pack_start (GTK_BOX (hbox), priv->power_switch, FALSE, FALSE, 4);
 
+	priv->visible_label = gtk_label_new (NULL);
+	gtk_misc_set_alignment (GTK_MISC (priv->visible_label), 0.0, 0.5);
+	gtk_label_set_line_wrap (GTK_LABEL (priv->visible_label), TRUE);
+	gtk_label_set_width_chars (GTK_LABEL (priv->visible_label), 30);
+	gtk_box_pack_start (GTK_BOX (vbox2), priv->visible_label, FALSE, FALSE, 4);
+
+	priv->visible_button = gtk_button_new_with_label (_("Make visible on Bluetooth"));
+	g_signal_connect (priv->visible_button, "clicked", G_CALLBACK (visible_button_cb), self);
+	gtk_widget_show (priv->visible_button);
+	gtk_box_pack_start (GTK_BOX (vbox2), priv->visible_button, FALSE, FALSE, 4);
+
 	/* Button for Send file */
 	priv->send_button = gtk_button_new_with_label (_("Send file from your computer"));
 	gtk_widget_show (priv->send_button);
@@ -1386,9 +1622,11 @@ create_devices_page (MoblinPanel *self)
                     G_CALLBACK (send_file_button_clicked_cb), self);
 	g_signal_connect (priv->display, "selected-device-changed",
 			G_CALLBACK (selected_device_changed_cb), self);
-	gtk_box_pack_start (GTK_BOX (vbox), priv->send_button, FALSE, FALSE, 4);
+	gtk_box_pack_start (GTK_BOX (vbox2), priv->send_button, FALSE, FALSE, 4);
 
-	enable_send_file (self);
+        powerswitch_state_changed_cb (priv->powerswitch,
+                                      bluetooth_powerswitch_get_state (priv->powerswitch),
+                                      self);
 
 	return page;
 }
@@ -1403,7 +1641,7 @@ create_connecting_page (MoblinPanel *self)
 
 	priv = MOBLIN_PANEL_GET_PRIVATE (self);
 
-	page = nbtk_gtk_frame_new ();
+	page = mx_gtk_frame_new ();
 	page_title = gtk_label_new ("");
 	gtk_frame_set_label_widget (GTK_FRAME (page), page_title);
 	//set_frame_title (GTK_FRAME (page), _("Connecting"));
@@ -1412,7 +1650,7 @@ create_connecting_page (MoblinPanel *self)
 	vbox = gtk_vbox_new (FALSE, 6);
 	gtk_widget_show (vbox);
 	gtk_container_add (GTK_CONTAINER (page), vbox);
-	priv->spinner = bling_spinner_new ();
+	priv->spinner = gtk_spinner_new ();
 	gtk_widget_set_size_request (priv->spinner, 150, 150);
 	gtk_widget_show (priv->spinner);
 	gtk_box_pack_start (GTK_BOX (vbox), priv->spinner, FALSE, FALSE, 6);
@@ -1462,6 +1700,7 @@ moblin_panel_init (MoblinPanel *self)
 {
 	MoblinPanelPrivate *priv;
 	GtkWidget *devices_page, *add_page, *setup_page, *ssp_setup_page, *failure_page, *connecting_page;
+	GtkWidget *page;
 
 	priv = MOBLIN_PANEL_GET_PRIVATE (self);
 	priv->pin_dialog = NULL;
@@ -1489,7 +1728,7 @@ moblin_panel_init (MoblinPanel *self)
 
 	bluetooth_agent_setup (priv->agent, AGENT_PATH);
 
-	priv->killswitch = bluetooth_killswitch_new ();
+	priv->powerswitch = bluetooth_powerswitch_new ();
 	bluetooth_plugin_manager_init ();
 
 	gtk_box_set_homogeneous (GTK_BOX (self), FALSE);
@@ -1519,6 +1758,10 @@ moblin_panel_init (MoblinPanel *self)
 	gtk_widget_show (connecting_page);
 	gtk_notebook_append_page (GTK_NOTEBOOK (priv->notebook), connecting_page, NULL);
 
+	page = create_summary_page (self);
+	gtk_widget_show (page);
+	gtk_notebook_append_page (GTK_NOTEBOOK (priv->notebook), page, NULL);
+
 	failure_page = create_failure_page (self);
 	gtk_widget_show (failure_page);
 	gtk_notebook_append_page (GTK_NOTEBOOK (priv->notebook), failure_page, NULL);
@@ -1535,9 +1778,20 @@ moblin_panel_dispose (GObject *object)
 
 	bluetooth_plugin_manager_cleanup ();
 
-	g_object_unref (priv->killswitch);
-	g_object_unref (priv->agent);
-	g_object_unref (priv->client);
+	if (priv->powerswitch != NULL) {
+		g_object_unref (priv->powerswitch);
+		priv->powerswitch = NULL;
+	}
+
+	if (priv->agent != NULL) {
+		g_object_unref (priv->agent);
+		priv->agent = NULL;
+	}
+
+	if (priv->client != NULL) {
+		g_object_unref (priv->client);
+		priv->client = NULL;
+	}
 
 	G_OBJECT_CLASS (moblin_panel_parent_class)->dispose (object);
 }
@@ -1554,6 +1808,12 @@ moblin_panel_class_init (MoblinPanelClass *klass)
 						G_STRUCT_OFFSET (MoblinPanelClass, status_connecting),
 						NULL, NULL, g_cclosure_marshal_VOID__BOOLEAN,
 						G_TYPE_NONE, 1, G_TYPE_BOOLEAN);
+
+	_signals[REQUEST_FOCUS] = g_signal_new ("request-focus", MOBLIN_TYPE_PANEL,
+						G_SIGNAL_RUN_FIRST,
+                                                0,
+						NULL, NULL, g_cclosure_marshal_VOID__VOID,
+						G_TYPE_NONE, 0);
 }
 
 /**
@@ -1569,12 +1829,19 @@ moblin_panel_new (void)
 
 /**
  *
- * moblin_panel_reset_view:
+ * moblin_panel_hidden:
  *
  * @self: A #MoblinPanel widget
  **/
 void
-moblin_panel_reset_view (MoblinPanel *self)
+moblin_panel_hidden (MoblinPanel *self)
 {
 	set_current_page (self, PAGE_DEVICES);
 }
+
+void
+moblin_panel_shown (MoblinPanel *self)
+{
+	set_current_page (self, PAGE_DEVICES);
+	make_discoverable (self);
+}
