@@ -42,10 +42,13 @@
 
 #define AGENT_PATH "/org/bluez/agent/sendto"
 
+#define RESPONSE_RETRY 1
+
 static DBusGConnection *conn = NULL;
+static ObexAgent *agent = NULL;
+static DBusGProxy *client_proxy = NULL;
 
 static GtkWidget *dialog;
-static GtkWidget *button;
 static GtkWidget *label_from;
 static GtkWidget *image_status;
 static GtkWidget *label_status;
@@ -55,6 +58,7 @@ static gchar *option_device = NULL;
 static gchar *option_device_name = NULL;
 static gchar **option_files = NULL;
 
+static DBusGProxy *current_transfer = NULL;
 static guint64 current_size = 0;
 static guint64 total_size = 0;
 static guint64 total_sent = 0;
@@ -65,6 +69,66 @@ static int file_index = 0;
 static gint64 first_update = 0;
 static gint64 last_update = 0;
 
+static void send_notify(DBusGProxy *proxy, DBusGProxyCall *call, void *user_data);
+
+/* Agent callbacks */
+static gboolean release_callback(DBusGMethodInvocation *context, gpointer user_data);
+static gboolean request_callback(DBusGMethodInvocation *context, DBusGProxy *transfer, gpointer user_data);
+static gboolean progress_callback(DBusGMethodInvocation *context,
+				  DBusGProxy *transfer,
+				  guint64 transferred,
+				  gpointer user_data);
+static gboolean complete_callback(DBusGMethodInvocation *context, DBusGProxy *transfer, gpointer user_data);
+static gboolean error_callback(DBusGMethodInvocation *context,
+			       DBusGProxy *transfer,
+			       const char *message,
+			       gpointer user_data);
+
+static void value_free(GValue *value)
+{
+	g_value_unset(value);
+	g_free(value);
+}
+
+static GHashTable *
+send_files (void)
+{
+	GHashTable *hash;
+	GValue *value;
+
+	hash = g_hash_table_new_full(g_str_hash, g_str_equal,
+				     g_free, (GDestroyNotify) value_free);
+
+	value = g_new0(GValue, 1);
+	g_value_init(value, G_TYPE_STRING);
+	g_value_set_string(value, option_device);
+	g_hash_table_insert(hash, g_strdup("Destination"), value);
+
+	dbus_g_proxy_begin_call(client_proxy, "SendFiles",
+				send_notify, NULL, NULL,
+				dbus_g_type_get_map("GHashTable", G_TYPE_STRING, G_TYPE_VALUE), hash,
+				G_TYPE_STRV, option_files,
+				DBUS_TYPE_G_OBJECT_PATH, AGENT_PATH,
+				G_TYPE_INVALID);
+
+	return hash;
+}
+
+static void
+setup_agent (void)
+{
+	if (agent == NULL)
+		agent = obex_agent_new();
+
+	obex_agent_set_release_func(agent, release_callback, NULL);
+	obex_agent_set_request_func(agent, request_callback, NULL);
+	obex_agent_set_progress_func(agent, progress_callback, NULL);
+	obex_agent_set_complete_func(agent, complete_callback, NULL);
+	obex_agent_set_error_func(agent, error_callback, NULL);
+
+	obex_agent_setup(agent, AGENT_PATH);
+}
+
 static gchar *filename_to_path(const gchar *filename)
 {
 	GFile *file;
@@ -125,11 +189,46 @@ static gchar *format_time(gint seconds)
 				"approximately %'d hours", hours), hours);
 }
 
+static void
+set_response_visible (GtkDialog *dialog,
+		      int response_id,
+		      gboolean visible)
+{
+	GtkWidget *widget;
+
+	widget = gtk_dialog_get_widget_for_response (dialog, response_id);
+	gtk_widget_set_no_show_all (widget, TRUE);
+	gtk_widget_set_visible (widget, visible);
+}
+
 static void response_callback(GtkWidget *dialog,
 					gint response, gpointer user_data)
 {
-	gtk_widget_destroy(dialog);
+	if (response == RESPONSE_RETRY) {
+		setup_agent ();
 
+		/* Reset buttons */
+		set_response_visible (GTK_DIALOG (dialog), GTK_RESPONSE_CLOSE, FALSE);
+		set_response_visible (GTK_DIALOG (dialog), RESPONSE_RETRY, FALSE);
+		set_response_visible (GTK_DIALOG (dialog), GTK_RESPONSE_CANCEL, TRUE);
+
+		/* Reset status and progress bar */
+		gtk_progress_bar_set_text (GTK_PROGRESS_BAR (progress),
+					  _("Connecting..."));
+		gtk_label_set_text (GTK_LABEL (label_status), "");
+		gtk_widget_hide (image_status);
+		send_files ();
+		return;
+	}
+
+	if (current_transfer != NULL) {
+		obex_agent_set_error_func(agent, NULL, NULL);
+		dbus_g_proxy_call_no_reply (current_transfer, "Cancel", G_TYPE_INVALID);
+		g_object_unref (current_transfer);
+		current_transfer = NULL;
+	}
+
+	gtk_widget_destroy(dialog);
 	gtk_main_quit();
 }
 
@@ -148,9 +247,13 @@ static void create_window(void)
 	gchar *text;
 
 	dialog = gtk_dialog_new_with_buttons(_("File Transfer"), NULL,
-				GTK_DIALOG_NO_SEPARATOR,
+				0,
+				GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
+				GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE,
+				_("_Retry"), RESPONSE_RETRY,
 				NULL);
-	button = gtk_dialog_add_button (GTK_DIALOG (dialog), GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL);
+	set_response_visible (GTK_DIALOG (dialog), GTK_RESPONSE_CLOSE, FALSE);
+	set_response_visible (GTK_DIALOG (dialog), RESPONSE_RETRY, FALSE);
 	gtk_window_set_type_hint(GTK_WINDOW(dialog),
 						GDK_WINDOW_TYPE_HINT_NORMAL);
 	gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
@@ -160,7 +263,8 @@ static void create_window(void)
 	vbox = gtk_vbox_new(FALSE, 0);
 	gtk_box_set_spacing(GTK_BOX(vbox), 6);
 	gtk_container_set_border_width(GTK_CONTAINER(vbox), 6);
-	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), vbox);
+	gtk_container_add (GTK_CONTAINER (gtk_dialog_get_content_area (GTK_DIALOG (dialog))),
+	                   vbox);
 
 	label = gtk_label_new(NULL);
 	gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
@@ -187,9 +291,6 @@ static void create_window(void)
 	label_from = gtk_label_new(NULL);
 	gtk_misc_set_alignment(GTK_MISC(label_from), 0, 0.5);
 	gtk_label_set_ellipsize(GTK_LABEL(label_from), PANGO_ELLIPSIZE_MIDDLE);
-	text = g_get_current_dir();
-	gtk_label_set_markup(GTK_LABEL(label_from), text);
-	g_free(text);
 	gtk_table_attach_defaults(GTK_TABLE(table), label_from, 1, 2, 0, 1);
 
 	label = gtk_label_new(NULL);
@@ -221,7 +322,7 @@ static void create_window(void)
 
 	label_status = gtk_label_new(NULL);
 	gtk_misc_set_alignment(GTK_MISC(label_status), 0, 0.5);
-	gtk_label_set_ellipsize(GTK_LABEL(label_status), PANGO_ELLIPSIZE_END);
+	gtk_label_set_line_wrap(GTK_LABEL(label_status), TRUE);
 	gtk_box_pack_start(GTK_BOX (hbox), label_status, TRUE, TRUE, 4);
 
 	gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 2);
@@ -237,7 +338,6 @@ static void create_window(void)
 static gchar *get_error_message(GError *error)
 {
 	char *message;
-	const gchar *name;
 
 	if (error == NULL)
 		return g_strdup(_("An unknown error occurred"));
@@ -247,8 +347,7 @@ static gchar *get_error_message(GError *error)
 		goto done;
 	}
 
-	name = dbus_g_error_get_name(error);
-	if (g_str_equal(name, OPENOBEX_CONNECTION_FAILED) == TRUE &&
+	if (dbus_g_error_has_name(error, OPENOBEX_CONNECTION_FAILED) == TRUE &&
 					is_palm_device(option_device)) {
 		message = g_strdup(_("Make sure that remote device "
 					"is switched on and that it "
@@ -256,7 +355,10 @@ static gchar *get_error_message(GError *error)
 		goto done;
 	}
 
-	message = g_strdup(error->message);
+	if (*error->message == '\0')
+		message = g_strdup(_("An unknown error occurred"));
+	else
+		message = g_strdup(error->message);
 
 done:
 	g_error_free(error);
@@ -311,6 +413,7 @@ static void get_properties_callback (DBusGProxy *proxy,
 {
 	GError *error = NULL;
 	gchar *filename = option_files[file_index];
+	GFile *file, *dir;
 	gchar *basename, *text, *markup;
 	GHashTable *hash;
 
@@ -329,9 +432,19 @@ static void get_properties_callback (DBusGProxy *proxy,
 		g_hash_table_destroy(hash);
 	}
 
-	text = g_path_get_dirname(filename);
-	gtk_label_set_markup(GTK_LABEL(label_from), text);
-	g_free(text);
+	file = g_file_new_for_path (filename);
+	dir = g_file_get_parent (file);
+	g_object_unref (file);
+	if (g_file_has_uri_scheme (dir, "file") != FALSE) {
+		text = g_file_get_path (dir);
+	} else {
+		text = g_file_get_uri (dir);
+	}
+	markup = g_markup_escape_text (text, -1);
+	g_free (text);
+	g_object_unref (dir);
+	gtk_label_set_markup(GTK_LABEL(label_from), markup);
+	g_free(markup);
 
 	basename = g_path_get_basename(filename);
 	text = g_strdup_printf(_("Sending %s"), basename);
@@ -350,10 +463,14 @@ static void get_properties_callback (DBusGProxy *proxy,
 static gboolean request_callback(DBusGMethodInvocation *context,
 				DBusGProxy *transfer, gpointer user_data)
 {
+	g_assert (current_transfer == NULL);
+
 	dbus_g_proxy_begin_call(transfer, "GetProperties",
 				get_properties_callback, NULL, NULL,
 				G_TYPE_INVALID);
 
+	current_transfer = g_object_ref (transfer);
+
 	dbus_g_method_return(context, "");
 
 	return TRUE;
@@ -425,7 +542,12 @@ static gboolean complete_callback(DBusGMethodInvocation *context,
 
 	file_index++;
 
-	gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(progress), 1.0);
+	/* And we're done with the transfer */
+	g_object_unref (current_transfer);
+	current_transfer = NULL;
+
+	if (file_index == file_count)
+		gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(progress), 1.0);
 
 	dbus_g_method_return(context);
 
@@ -437,6 +559,8 @@ static gboolean release_callback(DBusGMethodInvocation *context,
 {
 	dbus_g_method_return(context);
 
+	agent = NULL;
+
 	gtk_label_set_markup(GTK_LABEL(label_status), NULL);
 
 	gtk_widget_destroy(dialog);
@@ -446,6 +570,31 @@ static gboolean release_callback(DBusGMethodInvocation *context,
 	return TRUE;
 }
 
+static gboolean error_callback(DBusGMethodInvocation *context,
+			       DBusGProxy *transfer,
+			       const char *message,
+			       gpointer user_data)
+{
+	gtk_widget_show (image_status);
+	gtk_label_set_markup(GTK_LABEL(label_status), message);
+
+	set_response_visible (GTK_DIALOG (dialog), GTK_RESPONSE_CANCEL, TRUE);
+	set_response_visible (GTK_DIALOG (dialog), GTK_RESPONSE_CLOSE, FALSE);
+	set_response_visible (GTK_DIALOG (dialog), RESPONSE_RETRY, TRUE);
+
+	g_object_unref (current_transfer);
+	current_transfer = NULL;
+
+	if (agent != NULL) {
+		obex_agent_set_release_func(agent, NULL, NULL);
+		agent = NULL;
+	}
+
+	dbus_g_method_return(context);
+
+	return TRUE;
+}
+
 static void send_notify(DBusGProxy *proxy,
 				DBusGProxyCall *call, void *user_data)
 {
@@ -460,9 +609,9 @@ static void send_notify(DBusGProxy *proxy,
 		gtk_label_set_markup(GTK_LABEL(label_status), message);
 		g_free (message);
 
-		gtk_button_set_label (GTK_BUTTON (button), GTK_STOCK_CLOSE);
-		gtk_dialog_set_response_sensitive(GTK_DIALOG(dialog),
-						GTK_RESPONSE_CANCEL, TRUE);
+		set_response_visible (GTK_DIALOG (dialog), GTK_RESPONSE_CANCEL, TRUE);
+		set_response_visible (GTK_DIALOG (dialog), GTK_RESPONSE_CLOSE, FALSE);
+		set_response_visible (GTK_DIALOG (dialog), RESPONSE_RETRY, TRUE);
 		return;
 	}
 
@@ -471,12 +620,6 @@ static void send_notify(DBusGProxy *proxy,
 	first_update = get_system_time();
 }
 
-static void value_free(GValue *value)
-{
-	g_value_unset(value);
-	g_free(value);
-}
-
 static void
 select_device_changed(BluetoothChooser *sel,
 		      char *address,
@@ -491,12 +634,12 @@ select_device_changed(BluetoothChooser *sel,
 static char *
 show_browse_dialog (char **device_name)
 {
-	GtkWidget *dialog, *selector, *send_button, *image;
+	GtkWidget *dialog, *selector, *send_button, *image, *content_area;
 	char *bdaddr;
 	int response_id;
 
 	dialog = gtk_dialog_new_with_buttons(_("Select Device to Send To"), NULL,
-					     GTK_DIALOG_NO_SEPARATOR,
+					     0,
 					     GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT,
 					     NULL);
 	send_button = gtk_dialog_add_button (GTK_DIALOG (dialog), _("Send _To"), GTK_RESPONSE_ACCEPT);
@@ -507,7 +650,8 @@ show_browse_dialog (char **device_name)
 	gtk_window_set_default_size(GTK_WINDOW(dialog), 480, 400);
 
 	gtk_container_set_border_width (GTK_CONTAINER (dialog), 5);
-	gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->vbox), 2);
+	content_area = gtk_dialog_get_content_area (GTK_DIALOG (dialog));
+	gtk_box_set_spacing (GTK_BOX (content_area), 2);
 
 	selector = bluetooth_chooser_new("Select device to send files to");
 	gtk_container_set_border_width(GTK_CONTAINER(selector), 5);
@@ -519,7 +663,7 @@ show_browse_dialog (char **device_name)
 		     NULL);
 	g_signal_connect(selector, "selected-device-changed",
 			 G_CALLBACK(select_device_changed), dialog);
-	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), selector);
+	gtk_container_add (GTK_CONTAINER (content_area), selector);
 	bluetooth_chooser_start_discovery (BLUETOOTH_CHOOSER (selector));
 
 	bdaddr = NULL;
@@ -581,10 +725,6 @@ static GOptionEntry options[] = {
 
 int main(int argc, char *argv[])
 {
-	DBusGProxy *proxy;
-	GHashTable *hash = NULL;
-	GValue *value;
-	ObexAgent *agent;
 	GError *error = NULL;
 	int i;
 
@@ -691,36 +831,16 @@ int main(int argc, char *argv[])
 
 	create_window();
 
-	proxy = dbus_g_proxy_new_for_name(conn, "org.openobex.client",
+	client_proxy = dbus_g_proxy_new_for_name(conn, "org.openobex.client",
 					"/", "org.openobex.Client");
 
-	agent = obex_agent_new();
+	setup_agent ();
 
-	obex_agent_set_release_func(agent, release_callback, NULL);
-	obex_agent_set_request_func(agent, request_callback, NULL);
-	obex_agent_set_progress_func(agent, progress_callback, NULL);
-	obex_agent_set_complete_func(agent, complete_callback, NULL);
-
-	obex_agent_setup(agent, AGENT_PATH);
-
-	hash = g_hash_table_new_full(g_str_hash, g_str_equal,
-				     g_free, (GDestroyNotify) value_free);
-
-	value = g_new0(GValue, 1);
-	g_value_init(value, G_TYPE_STRING);
-	g_value_set_string(value, option_device);
-	g_hash_table_insert(hash, g_strdup("Destination"), value);
-
-	dbus_g_proxy_begin_call(proxy, "SendFiles",
-				send_notify, NULL, NULL,
-				dbus_g_type_get_map("GHashTable", G_TYPE_STRING, G_TYPE_VALUE), hash,
-				G_TYPE_STRV, option_files,
-				DBUS_TYPE_G_OBJECT_PATH, AGENT_PATH,
-				G_TYPE_INVALID);
+	send_files ();
 
 	gtk_main();
 
-	g_object_unref(proxy);
+	g_object_unref(client_proxy);
 
 	dbus_g_connection_unref(conn);
 
